"""Testing code generated by nbdev in unpackai/pdf.py"""
# Generated automatically from notebook nbs/51_pdf_utils.ipynb

from unpackai.pdf import *

# Test Cell
# For Test Cases (might have duplicate import because it will be in a dedicated file)
import re
from pathlib import Path
from typing import List

import pytest
from test_common.utils_4_tests import DATA_DIR, compare_strings
from test_utils import GITHUB_TEST_DATA_URL, check_connection_github
from unpackai import nlp

# Test Cell
GITHUB_TEST_PDF = f"{GITHUB_TEST_DATA_URL}/Deep%20learning.pdf"
LOCAL_TEST_PDF = DATA_DIR / "Deep learning.pdf"
LOCAL_TEST_TXT = DATA_DIR / "Deep learning.txt"

LOCAL_TEST_PDF_SMALL = DATA_DIR / "to_download.pdf"
LOCAL_TEST_TXT_SMALL = DATA_DIR / "to_download.txt"


@pytest.fixture(scope="session")
def local_textual():
    return TextualPDF.from_path(LOCAL_TEST_PDF)


@pytest.fixture(scope="session")
def test_pdf_as_txt():
    return Path(LOCAL_TEST_TXT).read_text(encoding="utf-8")


def cleanup_spaces(text: str) -> str:
    return re.sub(r"[\r\n\t\s]+", " ", text, flags=re.S)


class TestTextualPDF:
    def test_from_path(self, local_textual):
        """Test extract Textual of PDF from local file"""
        t = local_textual
        assert "Deep learning" in t.text, f"Text parsed:\n{t.text}"

    def test_from_path_not_pdf(self):
        """Test extract Textual of non-PDF file"""
        textual = TextualPDF.from_path(LOCAL_TEST_TXT)
        assert textual.text == nlp.Textual.from_path(LOCAL_TEST_TXT).text

    def test_from_path_pdf(self, local_textual):
        """Test extract Textual of PDF from local path using from_path_pdf"""
        textual = TextualPDF.from_path_pdf(LOCAL_TEST_PDF)
        assert textual.text == local_textual.text

    @pytest.mark.parametrize(
        "pdf, txt, encoding",
        [
            (LOCAL_TEST_PDF_SMALL, LOCAL_TEST_TXT_SMALL, None),
            (LOCAL_TEST_PDF, LOCAL_TEST_TXT, "utf-8"),
        ],
        ids=["small", "utf-8"],
    )
    def test_from_path_pdf_no_cleanup(self, pdf: Path, txt: Path, encoding):
        """Test extract Textual from PDF without cleanup"""
        content = cleanup_spaces(txt.read_text(encoding=encoding))

        textual = TextualPDF.from_path_pdf(pdf, cleanup=False)
        textual_txt = cleanup_spaces(textual.text)
        assert textual_txt == content, compare_strings(content, textual_txt)

    @pytest.mark.parametrize("pages", [[0], [0, 1], [0, 2], range(1, 4)])
    def test_from_path_pdf_pages(self, pages):
        """Test extract Textual from PDF with specific pages"""
        text = TextualPDF.from_path_pdf(LOCAL_TEST_PDF, page_numbers=pages).text
        for p in pages:
            assert f"Page {p+1}" in text, f"Page {p+1} not found in Textual: {text}"
        pages_extracted = re.findall(r"Page (\d+)", text)
        extra_pages = set(pages_extracted) - set(str(p + 1) for p in pages)
        assert extra_pages == set(), f"Extra pages extracted: {extra_pages}"

    def test_from_url(self, check_connection_github, local_textual):
        """Test extract Textual of PDF from URL"""
        textual = TextualPDF.from_url(GITHUB_TEST_PDF)
        assert textual.text == local_textual.text, f"URL text: {textual.text}"

    def test_from_url_pdf(self, check_connection_github, local_textual):
        """Test extract Textual of PDF from URL using from_url_pdf"""
        textual = TextualPDF.from_url_pdf(GITHUB_TEST_PDF)
        assert textual.text == local_textual.text, f"URL text: {textual.text}"

    # TODO: Add Tests with password



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/51_deploy_app.ipynb (unless otherwise specified).

__all__ = ['deploy_app', 'load_module', 'import_from_module', 'blackify', 'StreamlitApp', 'StreamlitAppCustom']

# Cell
import json
import requests
from dataclasses import dataclass, field
from importlib.machinery import SourceFileLoader
from importlib.util import spec_from_loader, module_from_spec
from pathlib import Path
from textwrap import dedent
from typing import Any, Union

from black import Mode, format_file_contents
from jinja2 import Template, DebugUndefined

from ..utils import IS_JUPYTER

# Internal Cell
PathStr = Union[Path, str]

# Cell
def deploy_app(app: PathStr = "app.py"):
    """Deploy a streamlit app using ngrok"""
    if not Path(app).is_file():
        print(f"ERROR: the app {app} does not exist!")
        return

    def show_url(public: str, port: str):
        print(f"Tunnel created for port {port} to: {public}")
        print("... click on the link to launch the app")
        if IS_JUPYTER:
            print("... Note: the output of streamlit is stored in nohup.out")

    if IS_JUPYTER:
        try:
            get_ipython().system_raw("ngrok http 8501 &")
            resp = requests.get("http://localhost:4040/api/tunnels")
        except Exception as e:
            print(f"Met error when trying to connect: {e}")
            print("🧙‍♂️This might happen: run the cell again ▶️ and it should work!")
            return

        tunnel = json.loads(resp.content)["tunnels"][0]
        local = tunnel["config"]["addr"]
        port = local.split(":")[-1]
        public = tunnel["public_url"]

        show_url(public, port)
        Path("nohup.out").write_text("")
        get_ipython().system_raw(f"nohup streamlit run {app}")
    else:
        raise NotImplementedError(f"Deployment outside jupyter currently not supported")


# Cell
def load_module(module_path: PathStr, module_name="module") -> Any:
    """Load and return a module"""
    module_path = Path(module_path)
    if not (module_path.is_file() and module_path.suffix == ".py"):
        raise AttributeError(f"Module {module_path} not found or not a .py file")
    loader = SourceFileLoader(module_name, str(module_path))
    spec = spec_from_loader(loader.name, loader)
    if not spec:
        raise AttributeError(f"Empty module spec from {module_path}")
    module = module_from_spec(spec)
    loader.exec_module(module)
    return module


def import_from_module(module_path: PathStr, name: str, module_name="module") -> Any:
    """Import a list of names from a module from its path"""
    module = load_module(module_path, module_name=module_name)
    try:
        return getattr(module, name)
    except AttributeError:
        raise AttributeError(f"Name {name} not found in {module_path}") from None


# Cell
def blackify(code:str) -> str:
    """Run "black" on a code passed as string"""
    return format_file_contents(code, mode=Mode(), fast=False)

# Internal Cell
TEMPLATE_FLOW = """
from pathlib import Path

import streamlit as st

{{ specific_import }}
{{ custom_import }}

st.set_page_config(page_title="ML deployment, by unpackAI", page_icon="🚀")
st.image("https://unpackai.github.io/unpackai_logo.svg")
st.title("{{ title }}")
st.write("*by {{ author }}*")
st.write("---")

{{ load_model }}

{{ post_process_code }}

{{ display_prediction }}

{{ input_2_prediction }}
"""


@dataclass
class TemplateCode:
    specific_import: str = ""
    load_model: str = ""
    display_prediction: str = ""
    input_2_prediction: str = """\
        select = st.sidebar.radio("How to load {{ input }}?", ["from file{% if multiple %}s{% endif %}", "from URL"])
        st.sidebar.write("---")

        if select == "from URL":
            url = st.sidebar.text_input("url")
            if url:
                display_prediction(url)

        else:
            {% if multiple -%}
            files = st.sidebar.file_uploader("Choose {{ input }}", accept_multiple_files=True)
            for file in files:  # type:ignore # this is an iterable
                display_prediction(file)
            {% else -%}
            file = st.sidebar.file_uploader("Choose {{ input }}")
            if file:
                display_prediction(file)
            {% endif %}
    """


# Internal Cell

# NOTE: We need re-implementation of elements defined during loading

TEMPLATE_CV_FASTAI = TemplateCode(
    specific_import="""
        {# The ✨ below is to go around import modification by nbdev #}
        from {# ✨ #}unpackai.deploy.cv import get_image, get_learner
    """
        ,
    load_model="""\
        learn = get_learner(Path(__file__).parent / "{{ model }}")
        vocab = learn.dls.vocab
    """,
    display_prediction="""
        def display_prediction(pic):
            img = get_image(pic)
            with learn.no_bar():
                prediction, idx, probabilities = learn.predict(img)
            col_img, col_pred = st.columns(2)
            col_img.image(img, caption=getattr(pic, "name", None))
            col_pred.write(f"### {prediction}")
            col_pred.metric(f"Probability", f"{probabilities[idx].item()*100:.2f}%")
    """,
)


# Internal Cell

# NOTE: Loading model and prediction depends on the module (e.g. "regression")

TEMPLATE_TABULAR_PYCARET = TemplateCode(
    specific_import="""
        import pandas as pd
        from pycaret.{{ module }} import load_model, predict_model
    """,
    load_model="""model = load_model("{{ model }}")""",
    display_prediction="""
        def display_prediction(csv):
            df = pd.read_csv(csv)
            predictions = predict_model(model, data = df)
            st.dataframe(predictions)
    """,
)


# Internal Cell
TEMPLATE_NLP_HF = TemplateCode(
    specific_import="""
        # TODO
    """,
    load_model="""
        # TODO
    """,
    display_prediction="""
        def display_prediction(csv:PathStr):
            pass # TODO
    """,
)


# Internal Cell

LIST_TEMPLATES = {
    "CV": {"fastai": TEMPLATE_CV_FASTAI},
    "Tabular": {"pycaret": TEMPLATE_TABULAR_PYCARET},
    "NLP": {"hugging_face": TEMPLATE_NLP_HF},
}

# Cell
@dataclass
class StreamlitApp:
    """Class to generate Streamlit App for different applications

    Available applications: "CV", "Tabular", "NLP", and "Custom"
    """

    application: str
    framework: str = field(init=False, default="")
    content: str = field(init=False, default="")
    _title: str = field(init=False, default="Streamlit App")
    _dest: Path = field(init=False, default_factory=Path)

    def __post_init__(self):
        applications = list(LIST_TEMPLATES)
        if self.application not in applications:
            raise AttributeError(
                f"ERROR: Application {self.application} not supported. "
                f"Available applications: {applications}"
            )

    @property
    def template(self) -> Template:
        """Get the template"""
        base_template = Template(TEMPLATE_FLOW, undefined=DebugUndefined)
        try:
            template_framework = LIST_TEMPLATES[self.application][self.framework]
        except KeyError as missing:
            raise AttributeError(f"Element {missing} not found among list of templates")

        return Template(
            base_template.render(
                specific_import=dedent(template_framework.specific_import),
                load_model=dedent(template_framework.load_model),
                display_prediction=dedent(template_framework.display_prediction),
                input_2_prediction=dedent(template_framework.input_2_prediction),
            )
        )

    def _render(
        self,
        framework: str,
        title: str,
        author: str,
        model: PathStr,
        post_process_code: str = "",
        custom_import: str = "",
        **kwargs,
    ):
        """Generic function for rendering"""
        self.framework = framework
        self._title = title
        input_ = {
            "CV": "images",
            "Tabular": "CSV",
            "NLP": "Text",
        }.get(self.application, "input")
        self.content = self.template.render(
            title=title,
            author=author,
            model=model,
            input=input_,
            multiple=(input_.endswith("s")),
            post_process_code=dedent(post_process_code),
            custom_import=dedent(custom_import),
            **kwargs,
        )
        self.content = blackify(self.content)
        return self

    def render_fastai(
        self,
        title: str,
        author: str,
        model: PathStr,
        post_process_code: str = "",
        custom_import: str = "",
    ) -> "StreamlitApp":
        """Render an app based on template

        Args:
            title: title of the App
            author: author of the App
            model: path of .pkl model to load (exported with `learn.export(...)`)
            post_process_mode: optional code to post-process the model after loading
            custom_import: optional code for custom import
        """
        return self._render(
            "fastai",
            title,
            author,
            model,
            post_process_code=post_process_code,
            custom_import=custom_import,
        )

    def render_pycaret(
        self,
        title: str,
        author: str,
        model: PathStr,
        module: str,
        post_process_code: str = "",
        custom_import: str = "",
    ) -> "StreamlitApp":
        """Render an app based on template

        Args:
            title: title of the App
            author: author of the App
            model: path of .pkl model to load (exported with `learn.export(...)`)
            module: regression, classification, etc.
            post_process_mode: optional code to post-process the model after loading
            custom_import: custom import (optional)
        """
        model = Path(model)
        return self._render(
            "pycaret",
            title,
            author,
            model.with_name(model.name.replace(".pkl", "")),
            module=module,
            post_process_code=post_process_code,
            custom_import=custom_import,
        )

    def append(self, content: str) -> "StreamlitApp":
        """Add additional content to the app"""
        self.content += dedent(content)
        return self

    def save(self, dest: PathStr = "app.py", show=False):
        """Write the app to a file"""
        Path(dest).write_text(self.content, encoding="utf-8")
        print(f"Saved app '{self._title}' to '{dest}'")
        if show:
            print("-" * 20)
            print(self.content)

    def deploy(self, dest: PathStr = "app.py") -> None:
        """Deploy the app"""
        if not self._dest:
            raise FileNotFoundError(f"App shall be saved before deployed")
        deploy_app(self._dest)


# Internal Cell
TEMPLATE_CUSTOM = TemplateCode(
    specific_import="{{ specific_import }}",
    load_model="",
    display_prediction="",
    input_2_prediction="{{ make_predictions }}"
)

# Cell
@dataclass
class StreamlitAppCustom:
    """Application with custom implementation"""

    content: str = field(init=False, default="")
    _title: str = field(init=False, default="Streamlit App")

    def render(
        self,
        title: str,
        author: str,
        make_predictions: Union[str, Path],
        custom_import: str = "",
    ) -> "StreamlitAppCustom":
        """Render an app based on template

        Args:
            title: title of the App
            author: author of the App
            make_predictions: custom implementations, either code or file path
                if a file is provided, it shall implement the function
                'def make_predictions()'
            custom_import: custom import (optional)
        """
        self._title = title
        base_template = Template(TEMPLATE_FLOW, undefined=DebugUndefined)
        template = Template(
            base_template.render(
                specific_import=dedent(TEMPLATE_CUSTOM.specific_import),
                load_model=dedent(TEMPLATE_CUSTOM.load_model),
                display_prediction=dedent(TEMPLATE_CUSTOM.display_prediction),
                input_2_prediction=dedent(TEMPLATE_CUSTOM.input_2_prediction),
            )
        )

        if isinstance(make_predictions, str) and not make_predictions.endswith(".py"):
            self.content = template.render(
                title=title,
                author=author,
                specific_import="",
                make_predictions=dedent(make_predictions),
                custom_import=custom_import,
            )
        elif isinstance(make_predictions, Path) or make_predictions.endswith(".py"):
            loaded_make_predictions = f"""\
                make_predictions = import_from_module(r"{make_predictions}", "make_predictions")
                make_predictions()
            """
            self.content = template.render(
                title=title,
                author=author,
                specific_import="from unpackai.deploy.app import import_from_module",
                make_predictions=dedent(loaded_make_predictions),
                custom_import=custom_import,
            )
        else:
            raise AttributeError(
                f"Incorrect type for {make_predictions}: shall be code or .py path"
            )

        self.content = blackify(self.content)
        return self

    def save(self, dest: PathStr, show=False):
        """Write the app to a file"""
        Path(dest).write_text(self.content, encoding="utf-8")
        print(f"Saved app '{self._title}' to '{dest}'")
        if show:
            print("-" * 20)
            print(self.content)


"""Testing code generated by nbdev in unpackai/deploy/app.py"""
# Generated automatically from notebook nbs/51_deploy_app.ipynb

from unpackai.deploy.app import *

# Test Cell
# For Test Cases (might have duplicate import because it will be in a dedicated file)
import re
from pathlib import Path
from typing import List

import pytest
from test_common.utils_4_tests import DATA_DIR, diff_files

DEPLOY_DATA_DIR = DATA_DIR / "deploy"

# Test Cell
def test_streamlit_cv_fastai(tmp_path):
    """Test Streamlit app for CV with fastai"""
    dest = tmp_path / "app_cv_fastai.py"
    StreamlitApp("CV").render_fastai(
        title="Is it a cat?", author="Jeff", model="model.pkl"
    ).append(
        """
        st.sidebar.write("---")
        st.sidebar.button("Show Balloons", on_click=st.balloons)
        """
    ).save(
        dest=dest
    )

    assert diff_files(DEPLOY_DATA_DIR / "app_CV.py", dest, show=True) == ""


# Test Cell
def test_streamlit_tabular_pycaret(tmp_path):
    """Test Streamlit app for Tabular with pycaret"""
    dest = tmp_path / "app_tabular_pycaret.py"
    StreamlitApp("Tabular").render_pycaret(
        title="My Mini Tabular App", author="Jeff", model="model.pkl", module="regression"
    ).save(dest=dest)

    assert diff_files(DEPLOY_DATA_DIR / "app_Tabular.py", dest, show=True) == ""


# Test Cell
def test_custom_app_code(tmp_path):
    """Test custom app with code provided via string"""
    dest = tmp_path / "app_custom_code.py"
    code = """
        st.write("hello")
        st.balloons()
    """
    StreamlitAppCustom().render("Custom App", "Jeff", code).save(dest=dest)

    assert diff_files(DEPLOY_DATA_DIR / "app_custom.py", dest, show=True) == ""


# Test Cell
def test_custom_app_external(tmp_path):
    """Test custom app with code provided via string"""
    dest = tmp_path / "app_custom_external.py"

    StreamlitAppCustom().render(
        "Custom App", "Jeff", DEPLOY_DATA_DIR / "correct_code_for_custom_app.py"
    ).save(dest=dest)

    # Because we have a absolute path of "test_data/deploy", we will replace by relative path
    # otherwise our tests will always be failed
    content = dest.read_text(encoding="utf-8").replace(
        str(DEPLOY_DATA_DIR), "<deploy_dir>"
    )
    content = content.replace(">\\", ">/").replace(
        'r"<deploy_dir>/', 'Path(__file__).parent / "'
    )
    dest.write_text(content, encoding="utf-8")

    assert diff_files(DEPLOY_DATA_DIR / "app_custom_external.py", dest, show=True) == ""



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/41_nlp_data.ipynb (unless otherwise specified).

__all__ = ['PathStr', 'Textual', 'HFTextBlock']

# Cell
import logging
import re
import requests
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import numpy as np
import pandas as pd
from bs4 import BeautifulSoup
from forgebox.html import DOM
from ipywidgets import interact, interact_manual, FileUpload
from unpackai import utils

PathStr = Union[Path, str]


# Cell
class Textual:
    """
    Obtain and manage textual data
    """

    def __init__(self, text: str, path: Optional[Path] = None):
        self.text = re.sub(r"[\r]?\n", " ", text)
        self.path = path

    def __repr__(self) -> str:
        return f"""Text ({len(self.text)} chars), textual(),
    train_path, val_path = textual.create_train_val()"""

    def __call__(self, page_size: int = 1000) -> None:
        """
        Previewing the first 200(or less) pages
        - page_size: character number each page
        """
        logging.info(f"Previewing the first 200 pages")

        @interact
        def show_text(page=(0, min(len(self.text)//page_size-1, 200), 1)):
            display(self.text[page*page_size: (page+1)*page_size])

    @classmethod
    def from_url(cls, url: str) -> 'Textual':
        """Create a Textual object from a URL"""
        return cls.from_path(utils.download(url))

    @classmethod
    def from_path(cls, path: PathStr) -> 'Textual':
        """
        Load a textual object from system path
        """
        path = Path(path)
        if not path.exists():
            raise FileExistsError(f"Cannot find {path}")
        return cls(path.read_text(encoding="utf-8"), path=path)

    @classmethod
    def from_upload(
        cls,
        path: Path = Path("./uploaded_file.txt")
    ):
        """
        Load textual with interactive upload button
        """
        DOM("🗃 Please upload a text file ended in .txt", "h4")()
        my_manual = interact_manual.options(manual_name="Upload")

        @my_manual
        def create_upload(btn_upload=FileUpload(description="Choose File")):
            text = list(btn_upload.values())[-1]['content'].decode()
            with open(path, "w") as f:
                f.write(text)
            return path

        def uploaded():
            result = create_upload.widget.result
            if result is None:
                raise FileExistsError(
                    "You have to upload the txt file first")
            return cls.from_path(result)
        return uploaded

    def create_train_val(
            self,
            valid_ratio=.2,
            train_path="./train_text.txt",
            val_path="./val_text.txt"):
        """
        create 2 files:
        - ./train_text.txt
        - ./val_text.txt
        """
        split = int(len(self.text)*(valid_ratio))
        with open(train_path, "w") as f:
            f.write(self.text[split:])
        with open(val_path, "w") as f:
            f.write(self.text[:split])
        return train_path, val_path

    def show_batch(self, tokenizer, bs:int = 4):
        from torch.utils.data.dataloader import DataLoader
        bunch = self.create_datasets(tokenizer)
        return next(iter(DataLoader(
            bunch['train_dataset'],
            batch_size=bs,
            collate_fn = bunch["data_collator"],
            )))

    def create_datasets(
        self,
        tokenizer,
        valid_ratio: float = .2,
        train_path: str = "./train_text.txt",
        val_path: str = "./val_text.txt",
        block_size: int = 128,
        mlm: bool = False,
    ) -> Dict[str, Any]:
        """
        Create pytorch datasets and collating functions
        - tokenizer: a huggingface tokenizer
        - valid ratio: portion of the valid data,
            compare to the entire dataset
        - train_path: a path saving train text file
        - val_path: a path saving valid text file
        - block_size: max possible length of the sequence
        - mlm, return a masked language modeling collating
            default False
        """
        # split dataset
        train_path, val_path = self.create_train_val(
            valid_ratio=valid_ratio,
            train_path=train_path,
            val_path=val_path,
        )

        from transformers import TextDataset, DataCollatorForLanguageModeling
        train_dataset = TextDataset(
            tokenizer=tokenizer,
            file_path=train_path,
            block_size=block_size)

        test_dataset = TextDataset(
            tokenizer=tokenizer,
            file_path=val_path,
            block_size=block_size)

        data_collator = DataCollatorForLanguageModeling(
            tokenizer=tokenizer, mlm=mlm,
        )
        return dict(
            train_dataset=train_dataset,
            eval_dataset=test_dataset,
            data_collator=data_collator,
        )

    def get_hf_trainer(
        self,
        model,
        tokenizer,
        arguments = None,
        valid_ratio: float = .2,
        train_path: str = "./train_text.txt",
        val_path: str = "./val_text.txt",
        block_size: int = 128,
        mlm: bool = False,
    ):
        """
        Create a hugging face trainer
        """
        from transformers import TrainingArguments, Trainer
        if arguments is None:
            arguments = TrainingArguments(
                output_dir="./write_style",
                overwrite_output_dir=True,  num_train_epochs=3,
                eval_steps = 400, save_steps=800, warmup_steps=600,
                per_device_train_batch_size=24,
                per_device_eval_batch_size=64,
            )
        trainer = Trainer(
            model=model, args=arguments,
            **self.create_datasets(tokenizer,
                valid_ratio=valid_ratio,
                train_path=train_path,
                val_path=val_path,
                block_size=block_size,
                mlm=mlm)
        )
        return trainer

# Cell


def HFTextBlock(tokenizer, **tk_kwargs):
    """
    create Huggingface specialized fastai Block
    tokenizer: PreTrainedTokenizer from huggingface
    tk_kwargs: keyward arguments for tokenizer's ```__call__```
    """
    from fastai.data.block import TransformBlock
    import torch

    def text_2_tensor_collate(data):
        """
        During the usual collation
        Use tokenizer to encode the text by batch
        """
        cols = list(zip(*data))
        result = []
        for col in cols:
            if type(col[0]) == str:
                result.append(
                    tokenizer(list(col), return_tensors="pt", **tk_kwargs)["input_ids"]
                )
            else:
                result.append(torch.stack(list(col), 0))
        return tuple(result)

    def get_hf_text_block():
        return TransformBlock(
            type_tfms=str, dls_kwargs={"create_batch": text_2_tensor_collate}
        )

    return get_hf_text_block



"""Testing code generated by nbdev in unpackai/cv/data.py"""
# Generated automatically from notebook nbs/25_cv_data.ipynb

from unpackai.cv.data import *

# Test Cell
# For Test Cases (might have duplicate import because it will be in a dedicated file)
from pathlib import Path
from PIL import Image
from tempfile import TemporaryDirectory
from typing import List

import pytest

from test_common.utils_4_tests import DATA_DIR, IMG_DIR
from test_utils import GITHUB_TEST_DATA_URL, check_connection_github

# Test Cell
tiff_fname = "tif-multi-frame-with-tags.tif"
gif_fname = "animated.gif"
LOCAL_TEST_TIFF = IMG_DIR / "tiff" / tiff_fname
LOCAL_TEST_GIF = IMG_DIR / "tiff" / gif_fname

LOCAL_TEST_SUMMARY_TAGS = DATA_DIR / "tiff_summary_tags.txt"
LOCAL_TEST_REPR = DATA_DIR / "tiff_repr.txt"

GITHUB_TEST_IMG_DIR_URL = (
    "https://raw.githubusercontent.com/unpackAI/unpackai/main/test/img"
)
GITHUB_TEST_TIFF = f"{GITHUB_TEST_IMG_DIR_URL}/tiff/{tiff_fname}"
GITHUB_TEST_GIF = f"{GITHUB_TEST_IMG_DIR_URL}/tiff/{gif_fname}"


def test_exception_on_missing_path():
    with pytest.raises(FileExistsError) as msg:
        TiffImage("blabla.tif")
    expected_msg = f"Cannot find path {'blabla.tif'}"
    assert str(msg.value) == expected_msg


def test_exception_on_wrong_suffix():
    with pytest.raises(ValueError) as msg:
        TiffImage(IMG_DIR / "tiff" / gif_fname)
    expected_msg = f"Image file should be .tif or .tiff, not '{'.gif'}'"
    assert str(msg.value) == expected_msg


@pytest.fixture(scope="session")
def test_tiff_image():
    """Fixture to pass the test TIFF image path"""
    return LOCAL_TEST_TIFF


@pytest.fixture(scope="session")
def test_repr():
    return LOCAL_TEST_REPR.read_text()


@pytest.fixture(scope="session")
def test_summary_tag_printout():
    return LOCAL_TEST_SUMMARY_TAGS.read_text()


@pytest.fixture(scope="session")
def local_TiffImage(test_tiff_image):
    """Create instance of TiffImage by loading the test TIFF image"""
    return TiffImage(test_tiff_image)


class Test_TiffImage:
    """Testing class for TiffImage"""

    def test_n_frames(self, local_TiffImage):
        """Test property n_frame"""
        expected_value = 8
        assert local_TiffImage.n_frames == expected_value

    def test_size(self, local_TiffImage):
        """Test property Size"""
        expected_value = (512, 512)
        assert local_TiffImage.size == expected_value

    def test_tags(self, local_TiffImage):
        """Test returned dictionary"""
        expected_value_0 = {
            256: 512,
            257: 512,
            259: 1,
            262: 1,
            270: '{"shape": [512, 512]}',
            273: (320,),
            277: 1,
            278: 512,
            279: (32768,),
            282: 1.0,
            283: 1.0,
            296: 1,
            65001: "Specular reflection",
            305: "tifffile.py",
        }
        expected_value_1 = {
            256: 512,
            257: 512,
            259: 1,
            262: 1,
            270: '{"shape": [512, 512]}',
            273: (33408,),
            277: 1,
            278: 512,
            279: (32768,),
            282: 1.0,
            283: 1.0,
            296: 1,
            65001: "Attached gingiva",
            305: "tifffile.py",
        }
        local_TiffImage.tiff.seek(0)
        assert local_TiffImage.tags == expected_value_0
        local_TiffImage.tiff.seek(1)
        assert local_TiffImage.tags == expected_value_1

    def test_repr(self, local_TiffImage, capsys, test_repr):
        """Test the __repr__"""
        expected_value = test_repr
        local_TiffImage.tiff.seek(0)
        print(local_TiffImage)
        captured = capsys.readouterr()
        assert captured.out == expected_value

    def test_summary_tags_printout(
        self, local_TiffImage, capsys, test_summary_tag_printout
    ):
        """Test the tag summary method"""
        expected_value = test_summary_tag_printout
        local_TiffImage.tiff.seek(0)
        local_TiffImage.summary_tags()
        captured = capsys.readouterr()
        assert captured.out == expected_value

    def test_type_returned_by_get_frame(self, local_TiffImage):
        """Test that object is 'PIL.TiffImagePlugin.TiffImageFile'"""
        frame = local_TiffImage.get_frame(0)
        assert isinstance(frame, TiffImageFile)

    def test_extract_one_frame(self, local_TiffImage):
        """Test that frames are extracted and saved"""
        with TemporaryDirectory() as tdir:
            p2dir = Path(tdir)
            tiff.extract_one_frame(frame_nbr=2, dest=p2dir)
            assert (p2dir / "tif-multi-frame-with-tags-0002.jpg").is_file()
            tiff.extract_one_frame(frame_nbr=3, dest=p2dir, image_format="tif")
            assert (p2dir / "tif-multi-frame-with-tags-0003.tif").is_file()
            tiff.extract_one_frame(frame_nbr=4, dest=p2dir, fname="specific-name")
            assert (p2dir / "specific-name.jpg").is_file()

    def test_extract_frames_by_tag(self, local_TiffImage):
        """Test that all frames are saved and naming by tag is correct"""
        with TemporaryDirectory() as tdir:
            p2dir = Path(tdir)
            tiff.extract_frames(dest=p2dir, naming_method="tag_value", tag=65001)
            assert (p2dir / "tif-multi-frame-with-tags-Hair.tif").is_file()
            assert (p2dir / "tif-multi-frame-with-tags-Skin.tif").is_file()
            assert (p2dir / "tif-multi-frame-with-tags-Stain.tif").is_file()
            assert (p2dir / "tif-multi-frame-with-tags-Marginal_gingiva.tif").is_file()
            assert len(list(p2dir.iterdir())) == local_TiffImage.n_frames

    def test_extract_frames_by_nbr(self, local_TiffImage):
        """Test that all frames are saved and naming by number is correct"""
        with TemporaryDirectory() as tdir:
            p2dir = Path(tdir)
            tiff.extract_frames(dest=p2dir)
            assert (p2dir / "tif-multi-frame-with-tags-0001.tif").is_file()
            assert (p2dir / "tif-multi-frame-with-tags-0002.tif").is_file()
            assert (p2dir / "tif-multi-frame-with-tags-0005.tif").is_file()
            assert (p2dir / "tif-multi-frame-with-tags-0007.tif").is_file()
            assert len(list(p2dir.iterdir())) == local_TiffImage.n_frames

    def test_is_valid_frame_(self, local_TiffImage):
        nframes = local_TiffImage.n_frames
        assert local_TiffImage.is_valid_frame_(0)
        with pytest.raises(ValueError) as msg:
            local_TiffImage.is_valid_frame_(nframes + 1)
        expected_msg = (
            f"'frame_nbr' is {nframes+1} but this TIFF file only has {nframes} frames."
        )
        assert str(msg.value)[:-1] == expected_msg



from difflib import unified_diff
from pathlib import Path
from typing import Union

import pytest
from pytest import LogCaptureFixture

TEST_DIR = Path(__file__).parent.parent
IMG_DIR = TEST_DIR / "img"
DATA_DIR = TEST_DIR / "test_data"

PathStr = Union[Path, str]


def compare_strings(
    exp: str,
    obt: str,
    name_exp: str = "Expected [string #1]",
    name_obt: str = "Obtained [string #2]",
) -> str:
    """Get unified diff of 2 string"""
    exp_lines = exp.splitlines(keepends=True)
    obt_lines = obt.splitlines(keepends=True)
    differences = unified_diff(exp_lines, obt_lines, fromfile=name_exp, tofile=name_obt)
    return "".join(differences)


def diff_files(exp: PathStr, obt: PathStr, encoding="utf-8", show=False) -> str:
    """Get unified diff of 2 files"""
    exp_str = Path(exp).read_text(encoding=encoding)
    obt_str = Path(obt).read_text(encoding=encoding)
    diff = compare_strings(exp_str, obt_str, name_exp=str(exp), name_obt=str(obt))
    if show:
        print(diff)
    return diff


def check_only_warning(caplog: LogCaptureFixture, sub_str: str):
    """Check that logging only contain warning and it contains a given substring"""
    assert caplog.records, f"There shall be at least one logged message"
    assert all(
        r.levelname == "WARNING" for r in caplog.records
    ), f"There shall be only warning (got {caplog.record_tuples})"
    assert sub_str in caplog.text, f"{sub_str} not found in {caplog.text}"


def check_no_log(caplog: LogCaptureFixture):
    """Check that there is nothing in logging"""
    assert (
        not caplog.records
    ), f"There should be no message but got {caplog.record_tuples}"



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/51_pdf_utils.ipynb (unless otherwise specified).

__all__ = ['PathStr', 'TextualPDF']

# Cell
import re
from pathlib import Path
from typing import List, Union

from pdfminer.high_level import extract_text
from unpackai import utils
from unpackai import nlp

PathStr = Union[Path, str]

# Cell
class TextualPDF(nlp.Textual):
    """Extend Textual for PDF"""

    @classmethod
    def from_url_pdf(
        cls, url: str, password: str = "", page_numbers: List[int] = None, cleanup=True
    ):
        """Create a Textual object from a PDF URL with specific options

        Args:
            url: url of PDF
            password: password, if the PDF is protected
            page_numbers: list of pages to extract (first page = 0)
            cleanup: remove messy characters and line returns (default=True)
        """
        return cls.from_path_pdf(
            utils.download(url),
            password=password,
            page_numbers=page_numbers,
            cleanup=cleanup,
        )

    @classmethod
    def from_path_pdf(
        cls,
        pdf_file: PathStr,
        password: str = "",
        page_numbers: List[int] = None,
        cleanup=True,
    ):
        """Create a Textual object from a PDF

        Args:
            pdf_file: path of PDF
            password: password, if the PDF is protected
            page_numbers: list of pages to extract (first page = 0)
            cleanup: remove messy characters and line returns (default=True)
        """
        txt = extract_text(pdf_file, password=password, page_numbers=page_numbers)
        if cleanup:
            txt = re.sub(r"[\r\n]{2,}", "<line_break>", txt)
            txt = re.sub(r"- *[\n\r]", "", txt)
            txt = txt.replace("\n", " ").replace("<line_break>", "\n\n")

        return cls(txt, Path(pdf_file))

    @classmethod
    def from_path(cls, path: PathStr):
        """Create a Textual object from a path, including PDF"""
        path = Path(path)
        if path.suffix.lower() == ".pdf":
            return cls.from_path_pdf(path)
        else:
            return super().from_path(path)



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/52_deploy_cv.ipynb (unless otherwise specified).

__all__ = ['get_learner', 'get_image', 'PathStr', 'dummy_function']

# Cell
import logging
import os
import pathlib
import re
import requests
import sys
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import List, Union, Tuple

import streamlit as st
from fastai.learner import load_learner, Learner
from fastai.vision.core import PILImage

# Internal Cell
logger = logging.getLogger("unpackai.deploy.cv")

# Internal Cell
@contextmanager
def set_posix():
    """To be able to load model in Windows"""
    posix_backup = pathlib.PosixPath
    try:
        if os.name == "nt":
            pathlib.PosixPath = pathlib.WindowsPath
        yield
    finally:
        pathlib.PosixPath = posix_backup


# Cell
PathStr = Union[Path, str]


@st.cache
def get_learner(model_path: PathStr) -> Learner:
    def try_loading(
        missing_implem: List[str] = None, nb_iter=5
    ) -> Tuple[List[str], Learner]:
        if missing_implem is None:
            missing_implem = []
        try:
            with set_posix():
                learner = load_learner(model_path)
        except AttributeError as e:
            m_missing_func = re.match(r"Can't get attribute '(.*?)'", str(e))
            if m_missing_func and nb_iter > 0:
                missing_implementation = m_missing_func.group(1)
                setattr(sys.modules["__main__"], missing_implementation, None)
                missing_implem.append(missing_implementation)
                return try_loading(missing_implem, nb_iter - 1)
            raise
        else:
            return missing_implem, learner

    missing_implem, learner = try_loading()
    if missing_implem:
        logger.warning(
            f"Missing function implementation: {missing_implem} => used 'None' instead"
        )

    return learner


@st.cache
def get_image(img: PathStr) -> PILImage:
    """Get picture from either a path or URL"""
    if str(img).startswith("http"):
        with tempfile.TemporaryDirectory() as tmpdirname:
            dest = Path(tmpdirname) / str(img).split("?")[0].rpartition("/")[-1]

            # NOTE: to be replaced by download(url, dest=dest) [from unpackai.utils]
            with requests.get(str(img)) as resp:
                resp.raise_for_status()
                dest.write_bytes(resp.content)

            return PILImage.create(dest)
    else:
        return PILImage.create(img)


# Cell
def dummy_function(*args, **kwargs):
    """Function that does absolutely nothing"""
    return None


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/25_cv_data.ipynb (unless otherwise specified).

__all__ = ['PathStr', 'TiffImage']

# Cell
# Import packages and modules required for cv.data module
import matplotlib.pyplot as plt
import numpy as np
import os
import string

from pathlib import Path
from PIL import Image, ImageSequence
from PIL.TiffImagePlugin import TiffImageFile, TiffTags, ImageFileDirectory_v2
from typing import Any, Dict, List, Optional, Union

PathStr = Union[Path, str]

# Cell
class TiffImage:
    """Load and handle TIFF images, such as viewing frames, extracting frames and saving them as single image files.

    TiffImage loads a TIFF image from a path (Path or str) and return a TiffImage object that:
    - gives access to the number of frames 'n_frames', all the tags 'tags' and the image size in pixel 'size'.
    - allows to return a specific frame as an object
    - allows to extract and save all or any frame as 'tif' or 'jpg image file(s)'
    - provides a __repr__ including information on each the frames in the image
    - allows to show thumbnails of all frames in a grid.

    Abreviations in the code are according to https://docs.fast.ai/dev/abbr.html

    """

    VALID_CHARS = f"-_.() {string.ascii_letters}{string.digits}"

    def __init__(self, path: PathStr):
        """Loads the TIFF image from system path"""
        self.path = self.handle_pathstr_(path)
        if self.path.suffix not in ['.tif', '.tiff']:
            raise ValueError(f"Image file should be .tif or .tiff, not '{self.path.suffix}'")
        self.tiff = TiffImageFile(self.path)

    def __repr__(self):
        """Return a summary of all the frames in the image file"""
        str_lst = [f"<unpackai.cv.data.TiffImage> TIFF image file with {self.n_frames} frames."]
        str_lst.append(f"  Loaded from {self.path}.")
        str_lst.append('  Frame Content Summary:')
        str_lst.append(f"    {'Frame':^7s}|{'Size':^15s}|{'Nbr Tags':^10s}")
        for i, frame in enumerate(ImageSequence.Iterator(self.tiff)):
            str_lst.append(f"    {str(i):^7s}|{str(frame.size):^15}|{str(len(frame.tag_v2.keys())):^10s}")
        str_lst.append(f"  To show image thumbnails, use method '.show(frame_nbr)' of '.show_all()'")
        txt = '\n'.join(str_lst)
        return txt

    def summary_tags(self, frame_nbr: int = 0):
        """Display a summary table of all the tags in the selected frame"""
        if self.is_valid_frame_(frame_nbr):
            self.tiff.seek(frame_nbr)
        tags_predefined = [tag for tag in self.tiff.tag_v2.keys() if tag in list(TiffTags.TAGS_V2.keys())]
        tags_custom = [tag for tag in self.tiff.tag_v2.keys() if tag not in list(TiffTags.TAGS_V2.keys())]

        str_lst = [f"|{'Tag Nbr':^9s}|{'Predefined Tag Name':^31s}|{'Tag Value':^70s} |"]
        str_lst.append(f"|{'=' * 112} |")
        str_lst.extend([f"|{str(tag):^9s}| {TiffTags.TAGS_V2[tag].name:<30s}| {str(self.tiff.tag_v2.get(tag))[:70]:<70s}|" for tag in tags_predefined])
        str_lst.extend([f"|{str(tag):^9s}| {' ':^30s}| {str(self.tiff.tag_v2.get(tag))[:70]:<70s}|" for tag in tags_custom])
        print('\n'.join(str_lst))

    def show(self, frame_nbr: int = 0):
        """Display a frame as a thumbnail"""
        if self.is_valid_frame_(frame_nbr):
            print(f"Showing frame {frame_nbr} out of {self.n_frames}:")
            self.tiff.seek(frame_nbr)
            plt.figure(figsize=(2, 2))
            plt.imshow(self.tiff)
            plt.axis('off')
            plt.show()

    def show_all(self, n_max: Optional[int] = None):
        """Display all frames (up to n_max) as a grid of thumbnails"""
        if n_max is None:
            n_max = self.n_frames
            print(f"Showing all {self.n_frames} frames:")
        else:
            n_max = min(n_max, self.n_frames)
            print(f"Showing {n_max} first frames of {self.n_frames}:")

        ncols = 6
        nrows = n_max//ncols + 1
        plt.figure(figsize=(14, 2 * nrows))
        for i in range(n_max):
            self.tiff.seek(i)
            plt.subplot(nrows, ncols, i+1)
            plt.imshow(self.tiff, cmap='Greys_r')
            plt.axis('off')
        plt.show()

    def get_frame(self, frame_nbr: int = 0) -> TiffImageFile:
        """Return the frame specified by 'frame_nbr' as an object

        The frame object is returned as a 'PIL.TiffImagePlugin.TiffImageFile' object. Which inherits from
        'PIL.Image.Image' and can be handled as a normal PIL image.
        """
        if self.is_valid_frame_(frame_nbr):
            self.tiff.seek(frame_nbr)
            return self.tiff

    def extract_frames(self,
                       dest: Optional[PathStr] = None,
                       naming_method: str = 'counter',
                       tag: Optional[int] = None
                      ):

        """Extract each frame from the file and save them as individual TIFF image file

        All frames in the TIFF file are saved as independant TIFF image files.

        Two options for the individual image files:
         - 'counter':    original file + frame number suffix (frame number is 4-digit padded with zeros)
         - 'tag_value':  use the value in a specific tag, typycally when a class name is stored in a tag for each frame.
                         In case a `tag` value is not provided, the naming will revert to `counter`.
                         In case no tag is availaible in the frame for `tag`, the naming for that particular frame will
                         use the counter value instead for that frame.
        """
        if tag is None:
            # todo: add warning here
            naming_method = 'counter'

        dest = self.handle_pathstr_(dest)

        for count, frame in enumerate(ImageSequence.Iterator(self.tiff)):
            ifd = frame.ifd
            if naming_method == 'counter':
                frame_slug = f"{count:04d}"
            elif naming_method == 'tag_value':
                if tag in list(frame.tag_v2.keys()):
                    tag_value = frame.tag_v2.get(tag)
                    frame_slug = ''.join(c for c in tag_value if c in self.VALID_CHARS).replace(' ', '_')
                else:
                    frame_slug = f"{count:04d}"
            else:
                raise ValueError(f"'naming_method' should be 'counter' or 'tag_value'")

            tiffpath = dest/f"{self.path.stem}-{frame_slug}.tif"
            frame.save(tiffpath, tiffinfo=ifd)

    def extract_one_frame(self,
                          frame_nbr: int = 0,
                          image_format: str = 'jpg',
                          dest: Optional[PathStr] = None,
                          fname: Optional[str] = None,
                         ):
        """Extract the specified frame and save it as an image file of the specified format

        image_format: str   'tif' to save as a TIFF image, along with the frame IFD (tags)
                            'jpg' to save as a JPEG image, loosing the tag information
        dest: Path or str   destination directory where to save the image file
                            when not provided, the current working directory is selected by default
        fname: str          specific name for the image file
                            when not provided, the name will be the same as the original TIFF image file
                            name, with the frame number as sufix (4-digit padded with zeroes)

        """

        if image_format not in ['jpg', 'tif']:
            raise ValueError(f"'image_format' must be 'jpg' or 'tif', not '{image_format}'")

        dest = self.handle_pathstr_(dest)

        if self.is_valid_frame_(frame_nbr):
            self.tiff.seek(frame_nbr)
            ifd = self.tiff.ifd
            frame_slug = f"{frame_nbr:04d}"
            if fname is None:
                fname = dest/f"{self.path.stem}-{frame_slug}.{image_format}"
            else:
                fname = dest/f"{fname}.{image_format}"

            self.tiff.save(fname, tiffinfo=ifd)

    @property
    def n_frames(self):
        """Number of frame in TIFF image file"""
        return self.tiff.n_frames

    @property
    def tags(self):
        """Dictionary with all tag-tag value pairs in the current frame"""
        return dict(self.tiff.tag_v2.items())

    @property
    def size(self):
        """Size of the current frame in pixel"""
        return self.tiff.size

    def is_valid_frame_(self, frame_nbr: int):
        """Internal utility method to validate that the requested frame exist for the current TIFF file"""
        if frame_nbr > self.n_frames:
            raise ValueError(f"'frame_nbr' is {frame_nbr} but this TIFF file only has {self.n_frames} frames. ")
        else:
            return True

    @staticmethod
    def handle_pathstr_(pathstr: PathStr):
        """Internal utility method to handle pathlib.Path or string as path and validate that the path exists"""
        if pathstr is None:
            pathstr = Path('')
        else:
            pathstr = Path(pathstr)
            if not pathstr.exists():
                raise FileExistsError(f"Cannot find path {pathstr}")
        return pathstr